// -*- C++ -*-
#if defined (ACE_HAS_VIRTUAL_TIME)

#include <ace/Thread_Manager.h>
#include <ace/VRT_Wrapper.h>

ACE_INLINE void 
ACE_Virtual_Real_Time::test_thr_mgr (ACE_Thread_Manager *tm)
{
  test_thr_mgr_ = tm;
}

ACE_INLINE ACE_Thread_Manager *
ACE_Virtual_Real_Time::test_thr_mgr (void)
{
  if (test_thr_mgr_ == 0) {
    test_thr_mgr_ = ACE_Thread_Manager::instance();
  }
  return test_thr_mgr_;
}

ACE_INLINE void 
ACE_Virtual_Real_Time::synchronizer (ACE_VRT_Synchronizer *s)
{
  if (delete_sync_) {
    delete synchronizer_;
  }
  delete_sync_ = 0;
  synchronizer_ = s;
}

ACE_INLINE ACE_VRT_Synchronizer *
ACE_Virtual_Real_Time::synchronizer (void)
{
  if (synchronizer_ == 0) {
    delete_sync_ = 1;
    synchronizer_ = new ACE_VRT_Synchronizer();
  }
  return synchronizer_;
}


ACE_INLINE ACE_thread_t
ACE_Virtual_Real_Time::owner(void)
{
  return owner_;
}

ACE_INLINE void
ACE_Virtual_Real_Time::set_owner(void)
{
  owner_ = ACE_Thread::self();
}

ACE_INLINE void
ACE_Virtual_Real_Time::set_secondary(void)
{
  secondary_ = ACE_Thread::self();
}

ACE_INLINE void 
ACE_Virtual_Real_Time::test_group (int x)
{
  test_group_ = x;
}

ACE_INLINE int 
ACE_Virtual_Real_Time::test_group (void)
{
  return test_group_;
}

ACE_INLINE int 
ACE_Virtual_Real_Time::in_test_group (void)
{
  ACE_thread_t tid = ACE_Thread::self();
  return vrt_enabled_ && (tid != owner_) && (tid != secondary_);
}

ACE_INLINE void 
ACE_Virtual_Real_Time::scale_factor (double df)
{
  scale_factor_ = df;
}

ACE_INLINE double 
ACE_Virtual_Real_Time::scale_factor (void)
{
  return scale_factor_;
}

ACE_INLINE void
ACE_Virtual_Real_Time::cpu_count(int c)
{
  cpu_count_ = c;
}

ACE_INLINE int
ACE_Virtual_Real_Time::cpu_count(void)
{
  return cpu_count_;
}

ACE_INLINE void
ACE_Virtual_Real_Time::synch_delay (const ACE_Time_Value &sd) 
{
  synch_delay_ = sd;
}

ACE_INLINE void 
ACE_Virtual_Real_Time::resume_test_threads(void)
{
  vrt_enabled_ = 1;
  if (suspended_time_ != ACE_Time_Value::zero)
    idle_time_ += (this->current_process_time() - suspended_time_);

  this->test_thr_mgr()->resume_grp(test_group_);

}

ACE_INLINE void 
ACE_Virtual_Real_Time::suspend_test_threads(void)
{
  suspended_time_ = this->current_process_time();
  this->test_thr_mgr()->suspend_grp(test_group_);
}

ACE_INLINE void 
ACE_Virtual_Real_Time::run_one_tick (long ticks)
{
  // this rather peculiar notation is used to ensure that NO MORE THAN
  // the desired number of clock ticks are scheduled. 
  struct timespec rqtp;

  rqtp.tv_sec = (ticks - 1) / 100;
  rqtp.tv_nsec = ((((ticks - 1) % 100))*10 + 5)*1000000;

  this->resume_test_threads();

  ::nanosleep(&rqtp,0);

  this->suspend_test_threads();
}


ACE_INLINE void
ACE_Virtual_Real_Time::release_test_threads (void)
{
  vrt_enabled_ = 0;
  this->test_thr_mgr()->resume_grp(test_group_);
}

ACE_INLINE ACE_Time_Value
ACE_Virtual_Real_Time::scaled_time_value (const ACE_Time_Value& tv)
{
  ACE_Time_Value result;
  double msec_rel = tv.msec() / (cpu_count_ * scale_factor_); 
  result.set(msec_rel/1000.0);
  return result;
}

ACE_INLINE ACE_Time_Value
ACE_Virtual_Real_Time::current_tick_time (void)
{
  ACE_Time_Value ctt = this->current_process_time() - idle_time_;
  double msec_rel = ctt.msec() * scale_factor_; 
  ctt.set(msec_rel/1000.0);
  return ctt;
} 

ACE_INLINE ACE_Time_Value
ACE_Virtual_Real_Time::current_process_time(void)
{
  tms buf;
  ::times(&buf);

  clock_t procticks = buf.tms_utime + buf.tms_stime - start_clock_time_;
  ACE_Time_Value result;
  result.set ((double)procticks / (double) CLK_TCK);
  
  return  result;
}

ACE_INLINE ACE_Time_Value
ACE_Virtual_Real_Time::current_virtual_time (void)
{
  return (vrt_enabled_) ? this->current_tick_time() + real_start_time_ 
    : ACE_VRT_Wrapper::getrealtimeofday();
}

// This routine does the housekeeping on time to wait.  It calculates
// a new value then checks what to do.  It returns a boolean to tell
// the caller if the wait is finished
ACE_INLINE int
ACE_Virtual_Real_Time::wait_longer (ACE_Time_Value expected_end_time, 
                                    ACE_Time_Value *time_to_wait)
{
  *time_to_wait =  expected_end_time - this->current_virtual_time();
  return (*time_to_wait > ACE_Time_Value::zero);
}


#endif /* ACE_HAS_VIRTUAL_TIME */
